#include "solver.hpp"
#include "common.hpp"

#include <iostream>
#include <cassert>
#include <algorithm>

using namespace sv;

using std::make_pair;
using std::pair;
using std::cout;
using std::endl;
using std::vector;

Model::Model():
	cn(0),
	bn(0),
	sense(0),
	vars(nullptr)
{

}

Var* Model::addVars(int num)
{
	Var* old = vars;
	vars = new Var[cn + num];
	for (int c = cn; c < cn + num; c++) {
		vars[c].col = c;
	}
	memcpy(vars, old, sizeof(Var) * cn);
	delete[] old;
	cn += num;
	return vars;
}

void Model::addConstr(const Expr& expr, ConstrOper sense, double rhs)
{
	if (sense == ConstrOper::LESS_EQUAL) {
		mt.push_back(vector<double>(1, rhs - expr.constant));
		mt.back().insert(mt.back().end(), expr.coeffs.begin(), expr.coeffs.end());
	}
	else if (sense == ConstrOper::GREATER_EQUAL) {
		mt.push_back(vector<double>(1, expr.constant - rhs));
		for (int coeff : expr.coeffs) {
			mt.back().push_back(-coeff);
		}
	}
	else {
		addConstr(expr, ConstrOper::LESS_EQUAL, rhs);
		addConstr(expr, ConstrOper::GREATER_EQUAL, rhs);
	}
	
	for (int c = mt.back().size(); c <= cn; c++) {
		mt.back().push_back(0);
	}
}

void Model::setObjective(Expr obje, int _sense)
{
	assert(_sense == 1 || _sense == -1);
	if (sense == 0) {
		mt.insert(mt.begin(), obje.coeffs);
		mt.front().insert(mt.front().begin(), -obje.constant);
		for (int c = obje.coeffs.size() + 1; c <= cn; c++) {
			mt.front().push_back(0);
		}
	}
	else {
		mt.front().front() = -obje.constant;
		for (int c = 0; c < cn; c++) {
			if (c < obje.coeffs.size()) {
				mt.front()[c + 1] = obje.coeffs[c];
			}
			else {
				mt.front()[c] = 0;
			}
		}
	}
	sense = _sense;
}

void Model::print()
{
	for (size_t i = 0; i < mt.size(); i++) {
		for (size_t j = 0; j < mt[0].size(); j++) {
			cout << mt[i][j] << "\t";
		}
		cout << endl;
	}
}

rtn Model::optimize()
{
	bn = mt.size();
	for (int row = 1; row < bn; row++) {
		mt.front().push_back(0);
		for (int col = 1; col < bn; col++) {
			mt[row].push_back(col == row ? 1 : 0);
		}
	}
	cn = mt.front().size();
	for (int i = 0; i < cn; i++) {
		mt.front()[i] = sense * mt.front()[i];
	}
	
	for (size_t i = 1; i < bn; i++) {
		basic.push_back(cn - bn + i);
	}

	// === жϳʼǷΪн ===
	bool initial_feasible = true;
	for (int row = 1; row < bn; row++) {
		if (mt[row].front() < 0) {
			initial_feasible = false;
			break;
		}
	}

	// === ʼн === 
	if (!initial_feasible) {
		vector<double> coeff = mt.front();
		mt.front() = vector<double>(cn, .0);
		mt.front().push_back(1);
		pair<size_t, size_t> t = { -1 ,cn };

		for (int row = 1; row < bn; row++) {
			mt[row].push_back(-1);
			if (t.first == -1 || mt[row].front() < mt[t.first].front()) {
				t.first = row;
			}
		}

		_gaussian(t);
		if (fabs(_simplex()) > 1e-10) {
			rtn_ = INFEASIBLE;
			return rtn_;
		}

		// if the x0 in B, we should pivot it.
		auto iter = find(basic.begin(), basic.end(), cn);
		if (iter != basic.end()) {
			for (int col = 1; col < mt.front().size(); col++) {
				if (fabs(mt.front()[col]) > 1e-10) {
					t = make_pair(iter - basic.begin() + 1, col);
					_gaussian(t);
					break;
				}
			}
		}

		for (int row = 0; row < bn; row++) {
			mt[row].pop_back();
		}

		// recover the coefficient line
		for (int col = 0; col < cn; col++) {
			mt.front()[col] = coeff[col];
		}
		for (int row = 1; row <= basic.size(); row++) {
			int norm = mt.front()[basic[row - 1]];
			for (int col = 0; col < cn; col++) {
				mt.front()[col] -= norm * mt[row][col];
			}
		}
	}

	res_ = -sense * _simplex();
	for (int row = 1; row < bn; row++) {
		if (basic[row - 1] - 1 < cn - bn) {
			vars[basic[row - 1] - 1].val = mt[row].front();
		}
	}
	return rtn_;
}

double Model::get(DoubleAttr attr)
{
	return res_;
}

int Model::get(IntAttr attr)
{
	return cn - bn;
}

double Model::_simplex()
{
	pair<size_t, size_t> t;
	while (1) {
		rtn_ = _pivot(t);
		if (rtn_ == OPTIMAL || rtn_ == UNBOUNDED) {
			break;
		}
		_gaussian(t);
	}
	return mt.front().front();
}

rtn Model::_pivot(pair<size_t, size_t>& p)
{
	p = make_pair(0, 0);
	double cmin = INT_MAX;
	vector<double> coef = mt.front();

	// === ԪСֵ ===
	for (size_t i = 1; i < coef.size(); i++) {
		if (cmin > coef[i] && find(basic.begin(), basic.end(), i) == basic.end()) {
			cmin = coef[i];
			p.second = i;
		}
	}
	if (cmin >= 0) {
		return OPTIMAL;
	}
	double bmin = INT_MAX;
	for (size_t row = 1; row < bn; row++) {
		double tmp = mt[row].front() / mt[row][p.second];
		if (mt[row][p.second] > 0 && bmin > tmp) {
			bmin = tmp;
			p.first = row;
		}
	}

	if (abs(bmin - INT_MAX) < 1e-10) {
		return UNBOUNDED;
	}

	for (auto iter = basic.begin(); iter != basic.end(); iter++) {
		if (mt[p.first][*iter] != 0) {
			*iter = p.second;
			break;
		}
	}
	assert(basic[p.first - 1] == p.second);
	return PIVOT;
}

void Model::_gaussian(pair<size_t, size_t> p)
{
	size_t x = p.first, y = p.second;

	// === йһ ===
	double norm = mt[x][y];
	for (size_t col = 0; col < mt[x].size(); col++) {
		mt[x][col] /= norm;
	}

	// === б任 ===
	for (size_t row = 0; row < bn; row++) {
		if (row == x) {
			continue;
		}
		if (mt[row][y] != 0) {
			double norm = mt[row][y];
			for (size_t col = 0; col < mt[x].size(); col++) {
				mt[row][col] = mt[row][col] - norm * mt[x][col];
			}
		}
	}
	basic[x - 1] = y;	// Ԫ
}

